/* ---------------------------------------------------------------------------
**	SDAC integrated PLL driver
** ---------------------------------------------------------------------------
**
**	Copyright	1997		Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** ---------------------------------------------------------------------------
**	This code was derived from the following sources of information:
**
**	[1]	SDAC Ramdac/Clock Synthesizer Data Book
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger
#define	DRIVER_NAME	"SDAC integrated PLL clock driver"
#define	DRIVER_REV	"$Revision: 1.3 $"

#include <ggi/module.h>
#include "S3/sdac.h"
#include "pll.h"

#define CLOCK_GENERATOR	3

static const struct pll_limits dclk_pll = 
{
	14318180,			/* fref				*/
	{ 1, 1 },			/* factor		 	*/
	{ 135000000, 270000000 },	/* fvco range			*/
	{ 0, 3 },			/* post-VCO divider range	*/
	{ 3, 129 },			/* dividend range		*/
	{ 3, 33 }			/* divisor range		*/
};

static struct kgi_clock clock =
{
	"S3 Incorporated",		/* manufacturer		*/
	"S3 86c716 (SDAC)",		/* model		*/
	GM_ALL,				/* supported modes 	*/
	0,				/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/

	CLK_TYPE_PROG,			/* clock type		*/
	{{
		25000000, 110000000,	/* clock range		*/
	}}
};

/*	A SDAC command word is as follows:
**
**	[0:6]	s->mul - 2	
**	[8:12]	s->div - 2
**	[13:14]	s->p
**	[16:24]	clock generator
*/

#define	INVALID_COMMAND	((ggi_uint) -1)

static ggi_uint pll_command(struct pll_state *s, struct kgi_mode *mode)
{
	ASSERT(s->mul - 2 < 128);
	ASSERT(s->div - 2 < 32);
	ASSERT(s->p < 4);

	return	(s->mul - 2) | ((s->div - 2) << 8) | (s->p << 13) |
		(CLOCK_GENERATOR << 16);
}

static void pll_set(struct kgi_display *dpy, ggi_uint cmd)
{
	/*	set PLL parameters
	*/
	DAC_out8(dpy, (cmd >> 16) & 7, DAC_PLL_W_INDEX);
	DAC_out8(dpy, cmd & 0xFF, DAC_PLL_DATA);
	DAC_out8(dpy, (cmd >> 8) & 0xFF, DAC_PLL_DATA);

	/*	external clock select lines
	*/
	DAC_out8(dpy, DAC_PLL_CONTROL, DAC_PLL_W_INDEX);
	DAC_out8(dpy, 0, DAC_PLL_DATA);

	/*	set clock select lines
	*/
	CLK_out8(dpy, (cmd >> 16) & 7, 0); 
}

/*
**	KGI module interface
*/

#define	NO_KGI_INIT_DONE_COMMAND
#include "pll.inc"

static uint16	pll_save[4];
static uint8	clk, pll_ctrl;

ggi_sint kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	clk = CLK_in8(dpy, 0);

	DAC_out8(dpy, 0, DAC_PLL_R_INDEX);
	DAC_ins8(dpy, DAC_PLL_DATA, pll_save, 8);

	NOTICE("please report the following values to %s: %.4x %.4x", \
		MAINTAINER, pll_save[0], pll_save[1]);

	DAC_out8(dpy, DAC_PLL_CONTROL, DAC_PLL_R_INDEX);
	pll_ctrl = DAC_in8(dpy, DAC_PLL_DATA);

	card->clock = &clock;
	return EOK;
}

void kgim_clock_done(struct kgi_display *dpy)
{
	CLK_out8(dpy, clk, 0);
	DAC_out8(dpy, 0, DAC_PLL_W_INDEX);
	DAC_outs8(dpy, DAC_PLL_DATA, pll_save, 8);

	DAC_out8(dpy, DAC_PLL_CONTROL, DAC_PLL_W_INDEX);
	DAC_out8(dpy, pll_ctrl, DAC_PLL_DATA);

	dpy->card->clock = NULL;
	NOTICE(DRIVER_NAME " removed");
}

ggi_sint kgim_clock_command(struct kgi_graphic *state, ggi_uint cmd, ggi_uint arg)
{
	return -E(CLOCK, UNKNOWN);
}

