/* --------------------------------------------------------------------------
**	OS and architecture dependent parts of GGI
** --------------------------------------------------------------------------
**
**	Copyright (C) 1997, Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
*/

#ifndef	_GGI_SYSTEM_H /* { */
#define	_GGI_SYSTEM_H

#include <ggi/debug.h>
#include <ggi/errno.h>

#define	KB	*(1024)
#define	MB	*(1024*1024)
#define	GB	*(1024*1024*1024)

/*	NOTE: This should make porting easier later. Porting KGI to another
**	NOTE: system would hopefully mean only to rewrite this header file
**	NOTE: once all OS specific stuff is in here. Currently, this is heavily
**	NOTE: Linux dependent.
*/

/* -----------------------------------------------------------------------------
**	Basic data types
** -----------------------------------------------------------------------------
**
**	boolean type
**		bool		TRUE: (value != 0), FALSE: (value == 0)
**
**	integer types
**		naming convention: (unsigned ? 'u' : 's') ## <base type>
**		where <base type> is one of the following
**
**		intf		fastest, depends on architecture >= 32 bit
**		intl		highest precision int, depends on architecture
**		int{8,16,32,64}	{8,16,32,64} bit integer
**
**	floating point types	(******!!!!!! DON'T USE IN KERNEL !!!!!!*******)
**		real4	4 byte IEEE floating point
**		real8	8 byte IEEE floating point 
**		real	fastest, depends on architecture
**		reall	highest precision, depends on architecture
*/

#if defined(linux) || defined(__linux) || defined(__linux__)
#include <linux/version.h>
#define	LinuxVersionCode(v, p, s) (((v)<<16)+((p)<<8)+(s))
#endif

#ifdef __KERNEL__ /* { */
#include <linux/kernel.h>
#include <linux/types.h>
#else 
#include <sys/types.h>
#endif /* } */

#ifndef __cplusplus
typedef int			bool;
#endif /* __cplusplus */

#if defined(linux) || defined(__linux) || defined(__linux__) /* { */

typedef signed char		sint8;
typedef unsigned char		uint8;

typedef signed short		sint16;
typedef unsigned short		uint16;

typedef signed int		sint32;
typedef unsigned int		uint32;

#if defined(GGI_NEED_64BIT)
typedef signed long long	sint64;
typedef unsigned long long	uint64;
#endif

typedef float			real4;
typedef double			real8;

#elif defined(__OpenBSD__) /* } { */

typedef int8_t			sint8;
typedef u_int8_t		uint8;

typedef int16_t			sint16;
typedef u_int16_t		uint16;

typedef int32_t			sint32;
typedef u_int32_t		uint32;

#if defined(GGI_NEED_64BIT)
typedef int64_t			sint64;
typedef u_int64_t		uint64;
#endif

#elif defined(sun) /* } { */

typedef int8_t			sint8;
typedef uint8_t			uint8;

typedef int16_t			sint16;
typedef uint16_t		uint16;

typedef int32_t			sint32;
typedef uint32_t		uint32;

typedef int64_t			sint64;
typedef uint64_t		uint64;

#elif defined(__alpha) /* } { */

typedef signed char		sint8;
typedef unsigned char		uint8;

typedef signed short		sint16;
typedef unsigned short		uint16;

typedef signed int		sint32;
typedef unsigned int		uint32;

typedef signed long		sint64;
typedef unsigned long		uint64;

typedef float			real4;
typedef double			real8;

#elif defined(__mips) && defined (__sgi) /* } { */

typedef signed char		sint8;
typedef unsigned char		uint8;

typedef signed short		sint16;
typedef unsigned short		uint16;

typedef signed int		sint32;
typedef unsigned int		uint32;

/* MIPS has several different sections. */
#if defined(_MIPS_SIM) /* { */

#if (_MIPS_SIM == _MIPS_SIM_NABI32) || (_MIPS_SIM == _MIPS_SIM_ABI32) /* { */

typedef signed long long	sint64;
typedef unsigned long long	uint64;

#elif _MIPS_SIM == _MIPS_SIM_ABI64 /* } { */

typedef signed long		sint64;
typedef unsigned long		uint64;

#else /* } { */
	#error Unknown MIPS architecture !
#endif /* } */
#else  /* } { */
	#error Compiler did not define MIPS architecture !
#endif /* } */

typedef float			real4;
typedef double			real8;

#elif defined(_AIX) /* } { */

typedef signed char		sint8;
typedef unsigned char		uint8;

typedef signed short		sint16;
typedef unsigned short		uint16;

typedef signed int		sint32;
typedef unsigned int		uint32;

#if defined(GGI_NEED_64BIT)
typedef signed long long	sint64;
typedef unsigned long long	uint64;
#endif

#else  /* } { */
	#error You need to edit system.h for your system!
#endif /* } */

#if defined(__i386__) ||\
 (defined(__sparc) || defined(sparc)) ||\
 (defined(__sgi) && defined (__mips)) ||\
 defined(_AIX) /* { */

typedef signed int		ggi_sint;
typedef unsigned int		ggi_uint;

#if defined(GGI_NEED_64BIT)
typedef signed long long	sintl;
typedef unsigned long long	uintl;
#else
typedef signed long		sintl;
typedef unsigned long		uintl;
#endif

typedef float			real;
typedef double			reall;

#elif defined (__alpha) /* } { */

typedef signed int		ggi_sint;
typedef unsigned int		ggi_uint;

typedef signed long		sintl;
typedef unsigned long		uintl;

typedef float			real;
typedef double			reall;

#else /* } { */
	#error	please define system dependent data types.
#endif /* } */

#define	BITS_PER(x)	(sizeof(x)*8)

#ifdef __KERNEL__

/*	A ..._region is used to describe a region in a physical address space.
**	Per physical address space, there are the following actions possible
**	with a region:
**
**	ggi_sint ..._check_region(struct ..._region *)
**		Before doing *any* access to a region, you have to check if the
**		region is not served by another driver. check_region() returns
**		EOK (== 0) if noone else claimed (a part) of this region before.
**		or a non-zero value otherwise.
**
**	..._vaddr ..._claim_region(struct ..._region *)
**		This makes a claim that you are going to handle this region
**		from now on. You must have checked a region successfully
**		before claiming it. A non-zero virtual address should be
**		returned.
**
**	..._vaddr ..._free_region(struct ..._region *)
**		Release a claim made earlier on this region. You must not
**		access the region after releasing your claim.
**
**	..._vaddr ..._alloc_region(struct ..._region *r)
**		Allocate and claim a free region of size r->size with r->decode
**		being the address lines decoded. This returns a non-zero
**		virtual address if successful or a zero value otherwise.
**
**	void ..._out<width>(?int<width> val, ..._vaddr vaddr)
**		Do the neccessary actions to write the value 'val' at
**		address 'addr'. 'addr' must be aligned to the <width> of
**		the value being written. The write cycle generated is of width
**		<width>.
**
**	ggi_uint<width> ..._in<width>(..._vaddr vaddr)
**		Do the neccessary actions to reat the cell at
**		address 'addr' which must be alligned to a cell width.
**		The read cycle generated is of width <width>.
*/

/*	Per region, we have a physical and a virtual address space and
**	certain information is stored with each region. As this is
**	common to all regions, let's make life easier.
*/

#define	DECLARE_IOSPACE(iospace, phys_t, virt_t)			  \
typedef phys_t iospace ## _addr; 	/* the physical address type	*/\
typedef virt_t iospace ## _vaddr; 	/* the virtual address type	*/\
struct iospace ## _region {						  \
	iospace ## _vaddr baseptr;	/* virtual base address		*/\
	iospace ## _addr base;		/* physical base address	*/\
	iospace ## _addr size;		/* size of region		*/\
	iospace ## _addr decode;	/* decoded phys address lines	*/\
	char *name;			/* name of the region		*/\
}

/*	Intel style in/out I/O space
**
**	The io_in.. / io_out.. functions generate the neccesary hardware actions
**	to do a read/write cycle in a cards I/O space. (Registers accessed via
**	the in.. / out.. instructions on i386 architecture.)
**
**	The function names encode direction (in == read cycle, out == write
**	cycle) and register width, which is 8, 16 or 32 bits. Any address given
**	__must__ be aligned to register widths, otherwise results may
**	be undefined!
*/

DECLARE_IOSPACE(io, uint16, uint16);
#define	IO_NULL		((io_vaddr) 0)

#include <asm/io.h>
#include <linux/ioport.h>

extern inline ggi_sint io_check_region(struct io_region *r)
{
	register ggi_sint foo = check_region(r->base, r->size);

	DEBUG2("checking io_region %s, base = %.4x, size %i, is %s", \
		r->name, r->base, r->size, foo ? "not free" : "free");
	return foo;
}

extern inline io_vaddr io_claim_region(struct io_region *r)
{
	PASSERT(r->baseptr == IO_NULL, "claim of claimed io_region '%s'!",\
		r->name);

	request_region(r->base, r->size, r->name);
	r->baseptr = r->base;

	DEBUG2("claiming io_region %s, base = %.4x, baseptr = %.4x", \
		r->name, r->base, r->baseptr);

	return r->baseptr;
}

extern inline io_vaddr io_free_region(struct io_region *r)
{
	DEBUG2("freeing io_region %s, base = %.4x, baseptr = %.4x", \
		r->name, r->base, r->baseptr);
	release_region(r->base, r->size);
	return (r->baseptr = IO_NULL);
}

extern inline io_vaddr io_alloc_region(struct io_region *r)
{
	/*	nothing appropriate under Linux. AFAIK - seeger
	*/
	return IO_NULL;
}

extern inline uint8 io_in8(io_vaddr addr) 
{
	return inb(addr);
}

extern inline uint16 io_in16(io_vaddr addr)
{
	return inw(addr);
}

extern inline uint32 io_in32(io_vaddr addr)
{
	return inl(addr);
}

extern inline void io_ins8(io_vaddr addr, void *buf, ggi_uint cnt)
{
	insb(addr, buf, cnt);
}

extern inline void io_ins16(io_vaddr addr, void *buf, ggi_uint cnt)
{
	insw(addr, buf, cnt);
}

extern inline void io_ins32(io_vaddr addr, void *buf, ggi_uint cnt)
{
	insl(addr, buf, cnt);
}

extern inline void io_out8(uint8 val, io_vaddr addr)
{
	outb(val, addr);
}

extern inline void io_out16(uint16 val, io_vaddr addr)
{
	outw(val, addr);
}

extern inline void io_out32(uint32 val, io_vaddr addr)
{
	outl(val, addr);
}

extern inline void io_outs8(io_vaddr addr, void *buf, ggi_uint cnt)
{
	outsb(addr, buf, cnt);
}

extern inline void io_outs16(io_vaddr addr, void *buf, ggi_uint cnt)
{
	outsw(addr, buf, cnt);
}

extern inline void io_outs32(io_vaddr addr, void *buf, ggi_uint cnt)
{
	outsl(addr, buf, cnt);
}

/*	memory I/O space
**
**	This is the 'normal' shared memory I/O space accessed using the
**	mov instructions on i386 architecture.
*/

DECLARE_IOSPACE(mem, uint32, void *);
#define	MEM_NULL	((mem_vaddr) 0)

#include <linux/mm.h>

extern inline ggi_sint mem_check_region(struct mem_region *r)
{
	/*	no ressource management for physical memory yet. AFAIK - seeger
	*/
	DEBUG2("checking mem_region %s, base %.8x, size %i, is always free!",\
		r->name, r->base, r->size);
	return EOK;
}

extern inline mem_vaddr mem_claim_region(struct mem_region *r)
{
	PASSERT(r->baseptr == MEM_NULL, "claim of claimed mem_region '%s'!", \
		r->name);

	if (r->base < 1 MB) {

		r->baseptr = phys_to_virt(r->base);

	} else {

		#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

			r->baseptr = vremap(r->base, r->size);
		#else
			r->baseptr = ioremap(r->base, r->size);
		#endif
	}

	DEBUG2("claiming mem_region %s, base = %.8x, baseptr = %p", \
		r->name, r->base, r->baseptr);
	return r->baseptr;
}

extern inline mem_vaddr mem_free_region(struct mem_region *r)
{
	PASSERT(r->baseptr, "free of freed region '%s'!", r->name);

	if (r->base >= 1 MB) {

		#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

			vfree(r->baseptr);
		#else
			iounmap(r->baseptr);
		#endif
	}

	DEBUG2("freeing mem_region %s, base = %.8x, baseptr = %p", \
		r->name, r->base, r->baseptr);
	return (r->baseptr = NULL);
}

extern inline mem_vaddr mem_alloc_region(struct mem_region *r)
{
	/*	no ressource management for physical memory yet. AFAIK - seeger
 	*/
	return MEM_NULL;
}

extern inline uint8 mem_in8(mem_vaddr addr)
{
	return 	readb(addr);
}

extern inline uint16 mem_in16(mem_vaddr addr)
{
	return readw(addr);
}

extern inline uint32 mem_in32(mem_vaddr addr)
{
	return readl(addr);
}

extern inline void mem_out8(uint8 val, mem_vaddr addr)
{
	writeb(val, addr);
}

extern inline void mem_out16(uint16 val, mem_vaddr addr)
{
	writew(val, addr);
}

extern inline void mem_out32(uint32 val, mem_vaddr addr)
{
	writel(val, addr);
}

#define	MEM_INS(x) \
extern inline void mem_ins##x(mem_vaddr addr, uint##x *buf, ggi_uint cnt)	\
{								\
	while (cnt--) {						\
								\
		*(buf++) = mem_in##x(addr);			\
	}							\
}

#define	MEM_OUTS(x) \
extern inline void mem_outs##x(mem_vaddr addr, uint##x *buf, ggi_uint cnt)	\
{								\
	while (cnt--) {						\
								\
		mem_out##x(*(buf++), addr);			\
	}							\
}

MEM_INS(8)
MEM_INS(16)
MEM_INS(32)

MEM_OUTS(8)
MEM_OUTS(16)
MEM_OUTS(32)

#undef	MEM_OUTS
#undef	MEM_INS

/*	PCI configuration space
**
**	These functions should be used to access the PCI configuration space.
**	As the address is given by the hardware wiring, a registration
**	scheme doesn't make sense. But I/O is pretty useful :)
*/

#include <linux/pci.h>
#include <linux/bios32.h>

typedef uint32 pcicfg_vaddr;
#define	PCICFG_NULL	((pcicfg_vaddr) -1)
#define	PCICFG_VADDR(bus, slot, fn)	\
	((pcicfg_vaddr)((((bus) & 0xFF) << 24) | (PCI_DEVFN(slot,fn) << 16)))

#define	PCIARGS	(addr >> 24) & 0xFF, (addr >> 16) & 0xFF, addr & 0xFF

#if CONFIG_PCI

extern inline uint8 pcicfg_in8(pcicfg_vaddr addr)
{
	uint8 tmp;
	pcibios_read_config_byte(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.2x\n", addr, tmp);
	return tmp;
}

extern inline uint16 pcicfg_in16(pcicfg_vaddr addr)
{
	uint16 tmp;
	pcibios_read_config_word(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.4x\n", addr,  tmp);
	return tmp;
}

extern inline uint32 pcicfg_in32(pcicfg_vaddr addr)
{
	uint32 tmp;
	pcibios_read_config_dword(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.8x\n", addr, tmp);
	return tmp;
}

extern inline void pcicfg_out8(uint8 val, pcicfg_vaddr addr)
{
	pcibios_write_config_byte(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.2x\n", addr, val);
}

extern inline void pcicfg_out16(uint16 val, pcicfg_vaddr addr)
{
	pcibios_write_config_word(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.4x\n", addr, val);
}

extern inline void pcicfg_out32(uint32 val, pcicfg_vaddr addr)
{
	pcibios_write_config_dword(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.8x\n", addr, val);
}

#undef PCIARGS

#else	/* #if CONFIG_PCI */

extern inline uint8 pcicfg_in8(pcicfg_vaddr addr)
{
	return -1;
}

extern inline uint16 pcicfg_in16(pcicfg_vaddr addr)
{
	return -1;
}

extern inline uint32 pcicfg_in32(pcicfg_vaddr addr)
{
	return -1;
}

extern inline void pcicfg_out8(uint8 val, pcicfg_vaddr addr)
{
}

extern inline void pcicfg_out16(uint16 val, pcicfg_vaddr addr)
{
}

extern inline void pcicfg_out32(uint32 val, pcicfg_vaddr addr)
{
}

#endif

#endif /* __KERNEL__ */

#endif	/* } #define _GGI_SYSTEM_H */
