/* 
 *  Graphics library for GGI. Mode management.
 *
 *  Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
 *
 *  If you do any modifications, I would like you to send diffs to me
 *  to allow for collecting a more and more complete set of drivers and
 *  to improve existing ones.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/ioctl.h>

#include "internal.h"
#include <ggi/ggi_commands.h>

extern int _ggiExtensionSize; 		/* init.c */
extern ggi_extension *_ggiExtension;

/* Open the dynamic libary requested
 */
static ggi_dlhandle *_ggiLoadDL(const char *filename)
{
	ggi_dlhandle hand,*hp;
#if defined(__OpenBSD__)
	char fn[1024];

	strcpy(fn,filename);
#endif

	hand.name=NULL;
	hand.use=0;
#if defined(__OpenBSD__)
	hand.handle=dlopen(fn,RTLD_LAZY);
#else
	hand.handle=dlopen(filename,RTLD_LAZY);
#endif
	DPRINT("hand.handle=%p\n",hand.handle);
	if (hand.handle==NULL) 
		return NULL;

	hand.init=(int (*)(struct ggi_visual *, const char *))  dlsym(hand.handle,GGI_DLINIT_SYM);
	hand.cleanup=(int (*)(struct ggi_visual *)) dlsym(hand.handle,GGI_DLCLEANUP_SYM);

	DPRINT("hand.init=%p\n",hand.init);
	DPRINT("hand.cleanup=%p\n",hand.cleanup);
	if (hand.init==NULL || hand.cleanup==NULL) {
		dlclose(hand.handle);
		return NULL;
	}

	hp=(ggi_dlhandle *)malloc(sizeof(ggi_dlhandle));
	memcpy(hp,&hand,sizeof(ggi_dlhandle));

	return hp;
}

/****** Open and Close a DL *********/
ggi_dlhandle *_ggiAddDL(ggi_visual *vis,const char *driver,
		const char *args,int type)
{
	const char *fname;
	ggi_dlhandle *dlh;
	ggi_dlhandle_l *tmp;
	int err;
	ggi_extension *ext;
	int size;

	fname=_ggiMatchConfig(driver,args);

	if (fname==NULL)
		return NULL;

	dlh=_ggiLoadDL(fname);
	DPRINT("%p=_ggiLoadDL(\"%s\")\n",dlh,fname);
	if (dlh==NULL)
		return NULL;
	
	err=dlh->init(vis,args);
	DPRINT("%d=dlh->init(%p,\"%s\") - %s %s\n",err,vis,args,driver,fname);
	if (err & GGI_DL_ERROR) 
		return NULL;
	dlh->type=type;
	if (type==GGI_DLTYPE_INTERNAL) {
		if (err & GGI_DL_OPDISPLAY) {
			tmp=(ggi_dlhandle_l *)malloc(sizeof(ggi_dlhandle_l));
			tmp->handle=dlh;
			tmp->next=vis->opdisplay->dlhandle;
			vis->opdisplay->dlhandle=tmp;
			dlh->use++;
		}
	
		if (err & GGI_DL_OPCOLOR) {
			tmp=(ggi_dlhandle_l *)malloc(sizeof(ggi_dlhandle_l));
			tmp->handle=dlh;
			tmp->next=vis->opcolor->dlhandle;
			vis->opcolor->dlhandle=tmp;
			dlh->use++;
		}

		if (err & GGI_DL_OP2D) {
			tmp=(ggi_dlhandle_l *)malloc(sizeof(ggi_dlhandle_l));
			tmp->handle=dlh;
			tmp->next=vis->op2d->dlhandle;
			vis->op2d->dlhandle=tmp;
			dlh->use++;
		}
		size=0;
		for (ext=_ggiExtension;ext && size<vis->size;ext=ext->next) {
			ext->load(vis,driver,args);
			size+=ext->size;
		}
	} else {
		dlh->use=1;
		tmp=(ggi_dlhandle_l *)malloc(sizeof(ggi_dlhandle_l));
		tmp->handle=dlh;
		tmp->next=vis->extlib;
		vis->extlib=tmp;
	}

	if (dlh->use==0) {
		fprintf(stderr,"libggi: %s (%s) -> 0x%.8x - no operations in this library\n",driver,args,err);
		dlclose(dlh->handle);
		free(dlh->name);
		free(dlh);
		return NULL;
	} else {
		tmp=(ggi_dlhandle_l *)malloc(sizeof(ggi_dlhandle_l));
		tmp->handle=dlh;
		tmp->next=VIS_DLHANDLE(vis);
		VIS_DLHANDLE(vis)=tmp;
	}

	dlh->name=strdup(driver);

	return dlh;
}

ggi_dlhandle *_ggiOpenDL(ggi_visual *vis,const char *driver,const char *args)
{
	return _ggiAddDL(vis,driver,args,GGI_DLTYPE_INTERNAL);
}

void _ggiRemoveDL(ggi_visual *vis,ggi_dlhandle_l **lib)
{
	ggi_dlhandle_l *tmp,**prev;
	ggi_dlhandle_l *libtmp,**libprev,*libnext;

	ggi_extension *ext,*last;
	int size;

	for (libprev=lib,libtmp=*lib;libtmp!=NULL; libtmp=libnext) {
		libnext=libtmp->next;
		if (libtmp->handle->use<=0) {
			DPRINT("Disposing \"%s\"\n",libtmp->handle->name);
			*libprev=libtmp->next;
			libtmp->handle->cleanup(vis);
			dlclose(libtmp->handle->handle);
			
			/* Now, clean up the master visual */
			prev=&VIS_DLHANDLE(vis);
			for (tmp=VIS_DLHANDLE(vis);tmp;tmp=tmp->next) {
				if (tmp->handle==libtmp->handle) break;
				prev=&tmp->next;
			}
			if (!tmp) DPRINT("Error: handle not in master list.\n");
			*prev=tmp->next;
			free(tmp);

			if (libtmp->handle->type!=GGI_DLTYPE_EXTENSION) {
				size=0;	
				ext=_ggiExtension;
				last=NULL;
				for (;ext && size<vis->size;ext=ext->next) {
					last=ext;
					size+=ext->size;
				}
				for (ext=last;ext;ext=ext->prev)
					ext->release(vis,libtmp->handle->name);
			}
			free(libtmp->handle->name);
			free(libtmp->handle);
			free(libtmp);
		} else {
			libprev=&libtmp->next;
		}
	}
}

static void _ggiZapDL(ggi_visual *vis,ggi_dlhandle_l **lib)
{
	ggi_dlhandle_l *tmp,*next;

	DPRINT("ZAPDL call: %p %p\n",vis,lib);

	for (tmp=*lib;tmp;tmp=tmp->next) 
		tmp->handle->use--;

	_ggiRemoveDL(vis,lib);

	for (tmp=*lib;tmp;tmp=next) {
		next=tmp->next;
		free(tmp);
	}

	*lib=NULL;
}

void _ggiCloseDL(ggi_visual *vis,ggi_uint type)
{
	if (vis==NULL)
		return;
	
	if ((type & GGI_DL_OPDISPLAY) && vis->opdisplay) {
		_ggiZapDL(vis,&vis->opdisplay->dlhandle);
	}

	if ((type & GGI_DL_OPCOLOR) && vis->opcolor) {
		_ggiZapDL(vis,&vis->opcolor->dlhandle);
	}

	if ((type & GGI_DL_OP2D) && vis->op2d) {
		_ggiZapDL(vis,&vis->op2d->dlhandle);
	}

	if ((type & GGI_DL_EXTENSION) && vis->extlib) {
		_ggiZapDL(vis,&vis->extlib);
	}

}
